from typing import *


# =====================================================================================================================
class Dicts(dict):  # use name *s to not mess with typing.Dict
    """
    just a class which keep all meths for dicts
    """
    def collapse_key(self, key: Any, source: dict = None) -> dict:
        """
        GOAL
        ----
        specially created for 2level-dicts (when values could be a dict)
        so it would replace values (if they are dicts and have special_key)

        CONSTRAINTS
        -----------
        it means that you have similar dicts with same exact keys
            {
                0: 0,
                1: {1:1, 2:2, 3:3},
                2: {1:11, 2:22, 3:33},
                3: {1:111, 2:222, 3:333},
                4: 4,
            }
        and want to get special slice like result

        SPECIALLY CREATED FOR
        ---------------------
        testplans get results for special dut from all results


        main idia to use values like dicts as variety and we can select now exact composition! remain other values without variants

        EXAMPLES
        --------
        dicts like
            {
                1: {1:1, 2:2, 3:3},
                2: {1:1, 2:None},
                3: {1:1},
                4: 4,
            }
        for key=2 return
            {
                1: 2,
                2: None,
                3: None,
                4: 4,
            }

        """
        result = {}
        if source is None:
            source = self

        for root_key, root_value in source.items():
            if isinstance(root_value, dict) and key in root_value:
                root_value = root_value.get(key)

            result[root_key] = root_value

        return result

    def prepare_serialisation(self, source: dict = None) -> dict:
        result = {}
        if source is None:
            source = self
        pass





        
        return result


# =====================================================================================================================



from typing import *
import pytest
from pytest import mark

from base_aux.base_argskwargs import *
from base_aux.funcs import *
from base_aux.objects import *


# =====================================================================================================================
def pytest_func_tester(
        func_link: TYPE__VALID_SOURCE, # if func would get Exx - instance of exx would be returned for value!
        args: TYPE__VALID_ARGS = Args(),
        kwargs: TYPE__VALID_KWARGS = Kwargs(),
        _EXPECTED: TYPE__VALID_RESULT = True,  # EXACT VALUE OR ExxClass

        # TODO: add validation func like in Valid!??
        _MARK: pytest.MarkDecorator | None = None,
        _COMMENT: str | None = None
) -> NoReturn | None:
    """
    NOTE
    ----
    this is same as funcs.Valid! except following:
        - if validation is Fail - raise assert!
        - no skips/cumulates/logs/ last_results/*values

    GOAL
    ----
    test target func with exact parameters
    no exception withing target func!

    TODO: apply Valid or merge them into single one!
    """
    if isinstance(args, ArgsKwargs):
        args = args.ARGS
    if isinstance(kwargs, ArgsKwargs):
        kwargs = kwargs.KWARGS

    if isinstance(args, NoValue):
        args = ()
    if isinstance(kwargs, NoValue):
        kwargs = dict()

    args = args__ensure_tuple(args)
    kwargs = kwargs or dict()
    comment = _COMMENT or ""

    if TypeChecker.check__callable_func_meth_inst(func_link):
        try:
            actual_value = func_link(*args, **kwargs)
        except Exception as exx:
            actual_value = exx
    else:
        actual_value = func_link

    print(f"pytest_func_tester={args=}/{kwargs=}//{actual_value=}/{_EXPECTED=}")

    # MARKS -------------------------
    # print(f"{mark.skipif(True)=}")
    if _MARK == mark.skip:
        pytest.skip("skip")
    elif isinstance(_MARK, pytest.MarkDecorator) and _MARK.name == "skipif" and all(_MARK.args):
        pytest.skip("skipIF")

    if _MARK == mark.xfail:
        if TypeChecker.check__exception(_EXPECTED):
            assert not TypeChecker.check__nested__by_cls_or_inst(actual_value, _EXPECTED), f"[xfail]{comment}"
        else:
            assert actual_value != _EXPECTED, f"[xfail]{comment}"
    else:
        if TypeChecker.check__exception(_EXPECTED):
            assert TypeChecker.check__nested__by_cls_or_inst(actual_value, _EXPECTED)
        else:
            assert actual_value == _EXPECTED


# ---------------------------------------------------------------------------------------------------------------------
def pytest_func_tester__no_args_kwargs(
        func_link: TYPE__VALID_SOURCE,
        _EXPECTED: TYPE__VALID_RESULT = True,

        _MARK: pytest.MarkDecorator | None = None,
        _COMMENT: str | None = None
) -> NoReturn | None:
    """
    created specially for using inline operators like 'func_link=A>=B'

    CAREFUL
    -------
    BUT be careful cause of exceptions!
    recommended using pytest_func_tester__no_args instead with 'func_link=lambda: A>=B'!!!
    """
    pytest_func_tester(func_link=func_link, _EXPECTED=_EXPECTED, _MARK=_MARK, _COMMENT=_COMMENT)


# ---------------------------------------------------------------------------------------------------------------------
def pytest_func_tester__no_kwargs(
        func_link: TYPE__VALID_SOURCE,
        args: TYPE__VALID_ARGS,
        _EXPECTED: TYPE__VALID_RESULT = True,

        _MARK: pytest.MarkDecorator | None = None,
        _COMMENT: str | None = None
) -> NoReturn | None:
    """
    short variant in case of kwargs is not needed

    WHY IT NEED
    -----------
    params passed by pytest while parametrisation as TUPLE!!!! so you cant miss any param in the middle!
    """
    pytest_func_tester(func_link=func_link, args=args, _EXPECTED=_EXPECTED, _MARK=_MARK, _COMMENT=_COMMENT)


# ---------------------------------------------------------------------------------------------------------------------
def pytest_func_tester__no_args(
        func_link: TYPE__VALID_SOURCE,
        kwargs: TYPE__VALID_KWARGS,
        _EXPECTED: TYPE__VALID_RESULT = True,

        _MARK: pytest.MarkDecorator | None = None,
        _COMMENT: str | None = None
) -> NoReturn | None:
    """
    short variant in case of args is not needed
    """
    pytest_func_tester(func_link=func_link, kwargs=kwargs, _EXPECTED=_EXPECTED, _MARK=_MARK, _COMMENT=_COMMENT)


# =====================================================================================================================
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------
pass    # USAGE EXAMPLES ----------------------------------------------------------------------------------------------


def _func_example(arg1: Any, arg2: Any) -> str:
    return f"{arg1}{arg2}"


# =====================================================================================================================
@pytest.mark.parametrize(argnames="func_link", argvalues=[_func_example, ])
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED, _MARK, _COMMENT",
    argvalues=[
        # TRIVIAL -------------
        ((1, None), {}, "1None", None, "ok"),
        ((1, 2), {}, "12", None, "ok"),

        # LIST -----------------
        ((1, []), {}, "1[]", None, "ok"),

        # MARKS -----------------
        ((1, 2), {}, None, mark.skip, "skip"),
        ((1, 2), {}, None, mark.skipif(True), "skip"),
        ((1, 2), {}, "12", mark.skipif(False), "ok"),
        ((1, 2), {}, None, mark.xfail, "ok"),
        # ((1, 2), {}, "12", mark.xfail, "SHOULD BE FAIL!"),
    ]
)
def test__full_params(func_link, args, kwargs, _EXPECTED, _MARK, _COMMENT):     # NOTE: all params passed by TUPLE!!!! so you cant miss any in the middle!
    pytest_func_tester(func_link, args, kwargs, _EXPECTED, _MARK, _COMMENT)


# =====================================================================================================================
@pytest.mark.parametrize(argnames="func_link", argvalues=[int, ])
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        (("1", ), {}, 1),
        ("1", {}, 1),                   # ARGS - direct one value acceptable
        (("hello", ), {}, Exception),   # EXPECT - direct exceptions
    ]
)
def test__short_variant(func_link, args, kwargs, _EXPECTED):
    pytest_func_tester(func_link, args, kwargs, _EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(argnames="func_link", argvalues=[int, ])
@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        (("1", ), 1),
        ("1", 1),
        ("", ValueError),
        (("hello", ), Exception),
    ]
)
def test__shortest_variant(func_link, args, _EXPECTED):
    pytest_func_tester__no_kwargs(func_link, args, _EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="expression",
    argvalues=[
        ("1rc2") == "1rc2",
        ("1rc2") != "1rc1",

        ("1.1rc2") > "1.0rc1",
        ("1.1rc2") > "1.1rc0",
        ("1.1rc2.0") > "1.1rc2",

        # ("01.01rc02") > "1.1rc1",
        ("01.01rc02") < "1.1rd1",
    ]
)
def test__expressions(expression):
    pytest_func_tester__no_args_kwargs(expression)


# =====================================================================================================================
# @pytest.mark.parametrize(
#     argnames="args, _EXPECTED",
#     argvalues=[
#         ((1, 1),        (True, True, False)),
#         ((1, 2),        (False, False, True)),
#         ((LAMBDA_TRUE, True), (False, False, True)),
#
#         ((ClsEq(1), 1), (True, True, False)),
#         ((ClsEq(1), 2), (False, False, True)),
#         ((1, ClsEq(1)), (True, True, False)),
#         ((2, ClsEq(1)), (False, False, True)),
#
#         ((ClsEqRaise(), 1), (Exception, False, True)),
#         ((1, ClsEqRaise()), (Exception, False, True)),
#     ]
# )
# def test__compare_doublesided(args, _EXPECTED):
#     func_link = Valid.compare_doublesided_or_exx
#     pytest_func_tester__no_kwargs(func_link, args, _EXPECTED[0])
#
#     func_link = Valid.compare_doublesided__bool
#     pytest_func_tester__no_kwargs(func_link, args, _EXPECTED[1])
#
#     func_link = Valid.compare_doublesided__reverse
#     pytest_func_tester__no_kwargs(func_link, args, _EXPECTED[2])


# =====================================================================================================================



import pathlib
from base_aux.classes import Text

try:
    import git  # need try statement! if not installed git.exe raise Exx even if module was setup!!!
except:
    pass


# =====================================================================================================================
def git_get_last_commit_short_info_str(source_path=None):    # starichenko
    """
    return last commit short info
    for git repository found in passed path
    """
    git_obj = None

    if not source_path:
        source_path = pathlib.Path.cwd()

    try:
        git_obj = git.Repo(source_path)
    except Exception as exx:
        print(f"{exx!r}")

    if git_obj:
        committer = git_obj.head.object.committer

        try:
            branch = git_obj.active_branch.name
            branch = Text(source=branch).shortcut(maxlen=15)
        except Exception as exx:
            msg = f"GIT DETACHED HEAD - you work not on last commit on brange! {exx!r}"
            print(msg)
            branch = "*DETACHED_HEAD*"

        summary = git_obj.commit().summary
        summary = Text(source=summary).shortcut(maxlen=15)

        hexsha = git_obj.head.object.hexsha[0:8]

        committed_datetime = str(git_obj.head.object.committed_datetime)[0:19]

        result = f"{branch}/{summary}/{committer}/{hexsha}/{committed_datetime}"

    else:
        result = f"возможно GIT не установлен"

    git_mark = f"[git_mark//{result}]"
    print(f"{git_mark=}")
    return git_mark


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "private_values"
    KEYWORDS: list[str] = [
        "environs", "environment",
        "private",
        "rc", "ini", "csv"
                     "json"
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "update values into class attrs from OsEnvironment or Ini/Json File"
    DESCRIPTION_LONG: str = """
    Designed to use private data like username/pwd kept secure in OsEnvironment or Ini/Json-File for your several home projects at ones.  
    And not open it in public.  

    **CAUTION:**  
    in requirements for other projects use fixed version! because it might be refactored so you would get exception soon.
    """
    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        ["load values to instance attrs from",
         "Environment",
         "IniFile",
         "JsonFile",
         "CsvFile",
         "direct text instead of file",
         "direct dict instead of file",
         ],

        ["attr access",
         "via any lettercase",
         "by instance attr",
         "like dict key on instance", ],

        ["work with dict", "apply", "update", "preupdate"],

        "update_dict as cumulative result - useful in case of settings result",
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 6, 1)
    TODO: list[str] = [
        "add Lock param after load?"
    ]
    FIXME: list[str] = [
        "..."
    ]
    NEWS: list[str] = [
        "[pypi.init] add static =NOW IT IS NOT WORKING!!!"
    ]


# =====================================================================================================================



from typing import Any, Callable, NoReturn, Union
from .novalue import NoValue


# =====================================================================================================================
TYPE__LAMBDA_CONSTRUCTOR = Any | type[Any] | Callable[..., Any | NoReturn]
TYPE__LAMBDA_ARGS = tuple[Any, ...]
TYPE__LAMBDA_KWARGS = dict[str, Any]


# =====================================================================================================================
class ArgsKwargs:
    """
    GOAL
    ----
    idea to keep args and kwargs in appropriate form/one object without application (constructor or func).
    so we can uncovering in later.
    usage in test parametrisation.

    SPECIALLY CREATED FOR
    ---------------------
    ATC tests with using special param prefix="*"

    BEST PRACTICE
    -------------
    for item, expect in [
        (ArgsKwargs("get name"), "ATC"),
        (ArgsKwargs("test gnd", _timeout=5), "PASS"),
    ]:
        assert serialDevice.send(*item.ARGS, **item.KWARGS) == expect

    WHY NOT - 1=add direct __iter for args and smth like __dict for kwargs
    ----------------------------------------------------------------------
    and use then (*victim, **victim)
    NO - there are no __dict like dander method!
    but we can use ArgsKwargs(dict)!? - yes but it add all other methods!
        class Cls(dict):
            ARGS: tuple[Any, ...]
            KWARGS: dict[str, Any]

            def __init__(self, *args, **kwargs) -> None:
                super().__init__(**kwargs)
                self.ARGS = args
                self.KWARGS = kwargs

            def __iter__(self) -> Iterator[Any]:
                yield from self.ARGS

    so as result the best decision is (*item.ARGS, **item.KWARGS)
    and we could use this class as simple base for Lambda for example!
    """
    ARGS: TYPE__LAMBDA_ARGS = ()
    KWARGS: TYPE__LAMBDA_KWARGS = {}

    def __init__(self, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs

    def __bool__(self) -> bool:
        if self.ARGS or self.KWARGS:
            return True
        else:
            return False


# ---------------------------------------------------------------------------------------------------------------------
class Args(ArgsKwargs):
    """
    just a derivative to clearly show only Args is important
    """
    def __bool__(self) -> bool:
        if self.ARGS:
            return True
        else:
            return False

    # def __iter__(self):
    #     yield from self.ARGS
    # NOTE: dont use any danders! its too complicated! be simple and get access to self.ARGS!


class Kwargs(ArgsKwargs):
    """
    just a derivative to clearly show only KwArgs is important
    """
    # TODO: decide apply dict nesting or not! by now it seems more clear
    def __bool__(self) -> bool:
        if self.KWARGS:
            return True
        else:
            return False

    # def __iter__(self):
    #     yield from self.KWARGS


# =====================================================================================================================
# SEE SAME BUT DIFFERS: TYPE__LAMBDA_ARGS *
TYPE__VALID_ARGS = Union[NoValue, Any, tuple, "TYPE__EXPLICIT", ArgsKwargs, Args]   # dont use None! use clear Args()/NoValue
TYPE__VALID_KWARGS = Union[NoValue, dict[str, Any], ArgsKwargs, Kwargs]             # dont use None! use clear Kwargs()/NoValue


# =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT


# =====================================================================================================================
# TEMPLATE
# from .main import (
#     # BASE
#     EXACT_OBJECTS,
#
#     # AUX
#
#     # TYPES
#
#     # EXX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .strings import (
    ReqCheckStr_Base,
    ReqCheckStr_Os,
    ReqCheckStr_Arch,

    _GetattrClassmethod_Meta,

    TYPE__VALUES,
    TYPE__RESULT_BOOL,
    TYPE__RESULT_RAISE,

    Exx_RequirementCantGetActualValue,
    Exx_Requirement,
)
from .pkgs import (
    Packages,
    CmdPattern,

    PATTERNS_IMPORT,
    PATTERN_IMPORT__MULTY_COMMA,
    PATTERN_IMPORT__MULTY_COMMA_BRACKETS,
    PATTERN_IMPORT__FROM,
)
from .versions import (
    Version,
    CheckVersion,
    CheckVersion_Python,

    VersionBlock,

    TYPE__VERSION_ELEMENT,
    TYPE__VERSION_ELEMENTS,
    TYPE__SOURCE_BLOCKS,

    TYPE__VERSION_BLOCKS,
    TYPE__SOURCE_VERSION,

    Exx_VersionIncompatible,
    Exx_VersionIncompatibleBlock,
    Exx_VersionIncompatibleCheck,
)


# =====================================================================================================================



# import time
import datetime


# =====================================================================================================================
class DateTime:
    TIMESTRUCT: datetime.datetime = None

    def __init__(self, timestruct = None):
        if timestruct is not None:
            self.TIMESTRUCT = timestruct

    def get_active__timestruct(self, value: datetime.datetime = None) -> datetime.datetime:
        # TODO: make get_first_valid from container!
        if value is None:
            value = self.TIMESTRUCT

        if value is None:
            value = datetime.datetime.now()

        return value

    def get_str(self, pattern: str = "%Y%m%d_%H%M%S", add_ms: bool = None, timestruct: datetime.datetime = None) -> str:
        """
        GOAL
        ----
        use for filenames like dumps/reservations/logs

        SPECIALLY CREATED FOR
        ---------------------

        EXAMPLES
        --------
        %Y%m%d_%H%M%S -> 20241203_114845
        add_ms -> 20241203_114934.805854
        """
        if add_ms:
            pattern += f".%f"

        timestruct = self.get_active__timestruct(timestruct)
        result = timestruct.strftime(pattern)
        return result

    def get_str__date(self, add_weekday: bool = None, pattern: str = "%Y.%m.%d", timestruct: datetime.datetime = None) -> str:
        """
        EXAMPLES
        --------
        %Y.%m.%d -> 2024.12.03.Tue
        """
        if add_weekday:
            pattern += ".%a"
        result = self.get_str(pattern=pattern, timestruct=timestruct)
        return result

    def get_str__time(self, add_ms: bool = None, pattern: str = "%H:%M:%S", timestruct: datetime.datetime = None) -> str:
        """

        EXAMPLES
        --------
        12:09:53
        12:11:53.855764
        """
        result = self.get_str(pattern=pattern, add_ms=add_ms, timestruct=timestruct)
        return result


# =====================================================================================================================
if __name__ == '__main__':
    print(DateTime().get_str__time(True))


# =====================================================================================================================



from typing import Union, Any, Callable, NoReturn
from base_aux.base_argskwargs.novalue import TYPE__NOVALUE


# =====================================================================================================================
TYPE__VALID_EXCEPTION = Union[Exception, type[Exception]]
TYPE__VALID_callable_item = Callable[[...], Any | NoReturn]
TYPE__VALID_SOURCE = Union[
    Any,                                # as main idea! as already final generic
    Callable[[...], Any | NoReturn],    # as main idea! to get final generic
    # TYPE__VALID_EXCEPTION,              # fixme: hide? think no? we can pass in theory Exx! but why? for tests? why? we need it only in result! if need in tests Place it as Any!
    TYPE__NOVALUE
]
TYPE__VALID_RESULT = Union[
    Any,
    TYPE__VALID_EXCEPTION,  # as main idea! instead of raise
]
# BOOL --------------------------------
TYPE__VALID_SOURCE_BOOL = Union[
    Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
    bool,                               # as main idea! as already final generic
    Callable[[...], bool | Any | NoReturn],   # as main idea! to get final generic
    # TYPE__VALID_EXCEPTION,
    TYPE__NOVALUE
]
TYPE__VALID_RESULT_BOOL = Union[
    # this is when you need get only bool! raise - as False!
    bool,  # as main idea! instead of raise/exx
]
# FIXME: TODO: solve idea of BOOL!!! cant understand about Exx in here!
TYPE__VALID_RESULT_BOOL__EXX = Union[
    bool,
    TYPE__VALID_EXCEPTION,
]

# ---------------------------------------------------------------------------------------------------------------------
TYPE__VALID_VALIDATOR = Union[
    Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
    # Type,   # Class as validator like Valid? fixme
    TYPE__VALID_EXCEPTION,  # direct comparison
    Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
]


# =====================================================================================================================
TYPES_ELEMENTARY_SINGLE: tuple[type, ...] = (
    type(None), bool,
    str, bytes,
    int, float,
)
TYPES_ELEMENTARY_COLLECTION: tuple[type, ...] = (
    tuple, list,
    set, dict,
)
TYPES_ELEMENTARY: tuple[type, ...] = (*TYPES_ELEMENTARY_SINGLE, *TYPES_ELEMENTARY_COLLECTION, )
TYPE__ELEMENTARY = Union[*TYPES_ELEMENTARY]


# =====================================================================================================================



from typing import *

from base_aux.objects import TypeChecker
from base_aux.funcs import *
from base_aux.valid import ValidAux
from base_aux.base_argskwargs.argskwargs import TYPE__LAMBDA_CONSTRUCTOR, TYPE__LAMBDA_ARGS, TYPE__LAMBDA_KWARGS


# =====================================================================================================================
class EqValidator:
    """
    base object to make a validation by direct comparing with other object
    no raise
    """

    VALIDATE_LINK: TYPE__VALID_VALIDATOR
    EXPECTED: bool | Any
    ARGS: TYPE__LAMBDA_ARGS
    KWARGS: TYPE__LAMBDA_KWARGS

    def __init__(self, validate_link: TYPE__VALID_VALIDATOR, *args, **kwargs) -> None:
        self.VALIDATE_LINK = validate_link
        self.ARGS = args
        self.KWARGS = kwargs

    def __eq__(self, other) -> bool:
        other = ValidAux.get_result_or_exx(other)
        args = (other, *self.ARGS)
        expected = ValidAux.get_result_or_exx(self.VALIDATE_LINK, *args, **self.KWARGS)

        result = ValidAux.eq_doublesided__bool(other, expected)
        return result

    def __call__(self, other: Any) -> bool:
        return self == other


# =====================================================================================================================



# =====================================================================================================================
class Exx__AnnotNotDefined(Exception):
    """Exception in case of not defined attribute in instance
    """


class Exx__NumberArithm_NoName(Exception):
    pass


class Exx__GetattrPrefix(Exception):
    pass


class Exx__GetattrPrefix_RaiseIf(Exx__GetattrPrefix):
    pass


class Exx__ValueNotParsed(Exception):
    pass


class Exx__ValueUnitsIncompatible(Exception):
    pass


class Exx__IndexOverlayed(Exception):
    pass


class Exx__IndexNotSet(Exception):
    pass


class Exx__ItemNotExists(Exception):
    """
    not exists INDEX (out of range) or NAME not in defined values
    """
    pass


class Exx__StartOuterNONE_UsedInStackByRecreation(Exception):
    """
    in stack it will be recreate automatically! so dont use in pure single BreederStrSeries!
    """
    pass


class Exx__BreederObjectList_GroupsNotGenerated(Exception):
    pass


class Exx__BreederObjectList_GroupNotExists(Exception):
    pass


class Exx__BreederObjectList_ObjCantAccessIndex(Exception):
    pass


# =====================================================================================================================
